# The name of the project
PROJECT_NAME = "paka"

# The service account that has access to all resources
ACCESS_ALL_SA = "access-all-sa"

# The environment variable for the directory where the paka data is saved
HOME_ENV_VAR = "PAKA_HOME"

# The environment variable for the buildpack builder
BP_BUILDER_ENV_VAR = "BP_BUILDER"

# The path where the model files are mounted in the container
MODEL_MOUNT_PATH = "/data"
